"""
Gaussian Quadratures Rules
--------------------------

:ref:`gaussian`
    The classical Gaussian quadrature scheme applied on any probability
    distribution.
:ref:`gauss_legendre`
    Same as :ref:`gaussian` for uniform distribution, but applicable to other
    distribution by incorporating the probability density as part of the
    function to be integrated.
:ref:`gauss_patterson`
    Extension of Gauss-Legendre rule. Valid to order 8.
:ref:`gauss_kronrod`
    Extension to the Gauss-Patterson rule to include most distribution and any
    order.
:ref:`gauss_lobatto`
    Gaussian quadrature rule that enforces the endpoints to be included in the
    rule.
:ref:`gauss_radau`
    Gaussian quadrature rule that enforces that a single fixed point to be
    included in the rule.

Non-Gaussian Quadrature Rules
-----------------------------

:ref:`clenshaw_curtis`
    Chebyshev nodes with endpoints included.
:ref:`fejer`
    Chebyshev nodes without endpoints included.
:ref:`leja`
    Fully nested quadrature method.
:ref:`genz_keister`
    Genz-Keizter 16 rule. Nested. Valid to order 8.
:ref:`newton_cotes`
    Numerical integration rule based on fixed width abscissas.
"""
import numpy

from .combine import combine

from .clenshaw_curtis import quad_clenshaw_curtis
from .fejer import quad_fejer
from .gaussian import quad_gaussian
from .gauss_patterson import quad_gauss_patterson
from .gauss_legendre import quad_gauss_legendre
from .gauss_lobatto import quad_gauss_lobatto
from .gauss_kronrod import quad_gauss_kronrod
from .gauss_radau import quad_gauss_radau
from .genz_keister import quad_genz_keister
from .leja import quad_leja
from .newton_cotes import quad_newton_cotes

QUAD_NAMES = {
    "c": "clenshaw_curtis", "clenshaw_curtis": "clenshaw_curtis",
    "f": "fejer", "fejer": "fejer",
    "g": "gaussian", "gaussian": "gaussian",
    "e": "gauss_legendre", "gauss_legendre": "gauss_legendre",
    "l": "gauss_lobatto", "gauss_lobatto": "gauss_lobatto",
    "k": "gauss_kronrod", "gauss_kronrod": "gauss_kronrod",
    "p": "gauss_patterson", "gauss_patterson": "gauss_patterson",
    "r": "gauss_radau", "gauss_radau": "gauss_radau",
    "z": "genz_keister", "genz_keister": "genz_keister",
    "j": "leja", "leja": "leja",
    "n": "newton_cotes", "newton_cotes": "newton_cotes",
}
QUAD_FUNCTIONS = {
    "clenshaw_curtis": quad_clenshaw_curtis,
    "fejer": quad_fejer,
    "gaussian": quad_gaussian,
    "gauss_kronrod": quad_gauss_kronrod,
    "gauss_legendre": quad_gauss_legendre,
    "gauss_lobatto": quad_gauss_lobatto,
    "gauss_patterson": quad_gauss_patterson,
    "gauss_radau": quad_gauss_radau,
    "genz_keister": quad_genz_keister,
    "leja": quad_leja,
    "newton_cotes": quad_newton_cotes,
}


def generate_quadrature(
        order,
        dist,
        rule="clenshaw_curtis",
        sparse=False,
        accuracy=100,
        growth=None,
        recurrence_algorithm="",
):
    """
    Numerical quadrature node and weight generator.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
        rule (str):
            Rule for generating abscissas and weights. Either done with
            quadrature rules, or with random samples with constant weights.
        sparse (bool):
            If True used Smolyak's sparse grid instead of normal tensor product
            grid.
        accuracy (int):
            If gaussian is set, but the Dist provieded in domain does not
            provide an analytical TTR, ac sets the approximation order for the
            descitized Stieltje's method.
        growth (bool, None):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples. Defaults to the same
            value as ``sparse`` if omitted.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights in
            case of Gaussian quadrature scheme. If omitted, ``analytical`` will
            be tried first, and ``stieltjes`` used if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Abscissas and weights created from full tensor grid rule. Flatten
            such that ``abscissas.shape == (len(dist), len(weights))``.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 3)
        >>> abscissas, weights = generate_quadrature(
        ...     1, distribution, rule=("gaussian", "fejer"))
        >>> print(abscissas)
        [[-1.   -1.    1.    1.  ]
         [-3.75  3.75 -3.75  3.75]]
        >>> print(weights)
        [0.25 0.25 0.25 0.25]
    """
    if sparse:
        from . import sparse_grid
        return sparse_grid.construct_sparse_grid(
            order, dist, rule=rule, accuracy=accuracy, growth=growth)

    if not isinstance(rule, str):
        order = numpy.ones(len(dist), dtype=int)*order
        abscissas, weights = zip(*[
            generate_quadrature(order_, dist_, rule_, growth)
            for order_, dist_, rule_ in zip(order, dist, rule)
        ])
        abscissas = combine([abscissa.T for abscissa in abscissas]).T
        weights = numpy.prod(combine([abscissa.T for abscissa in weights]), -1)
        return abscissas, weights

    rule = QUAD_NAMES[rule.lower()]
    kwargs = {}

    if rule in ("clenshaw_curtis", "fejer", "newton_cotes"):
        kwargs.update(growth=growth)

    if rule in ("gaussian", "gauss_kronrod", "gauss_radau", "gauss_lobatto"):
        kwargs.update(accuracy=accuracy,
                      recurrence_algorithm=recurrence_algorithm)

    quad_function = QUAD_FUNCTIONS[rule]
    abscissas, weights = quad_function(order, dist, **kwargs)

    assert len(weights) == abscissas.shape[1]
    assert len(abscissas.shape) == 2

    if getattr(dist, "interpret_as_integer", False):
        abscissas = abscissas.astype(int)

    return abscissas, weights



"""
Constructing custom probability distributions is done by subclassing the
distribution :class:`~chaospy.distributions.baseclass.Dist`::

    >>> class Uniform(Dist):
    ...     def __init__(self, lo=0, up=1):
    ...         '''Initializer.'''
    ...         Dist.__init__(self, lo=lo, up=up)
    ...     def _cdf(self, x_data, lo, up):
    ...         '''Cumulative distribution function.'''
    ...         return (x_data-lo)/(up-lo)
    ...     def _bnd(self, x_data, lo, up):
    ...         '''Lower and upper bounds.'''
    ...         return lo, up
    ...     def _pdf(self, x_data, lo, up):
    ...         '''Probability density function.'''
    ...         return 1./(up-lo)
    ...     def _ppf(self, q_data, lo, up):
    ...         '''Point percentile function.'''
    ...         return q_data*(up-lo) + lo

Usage is then straight forward::

    >>> dist = Uniform(-3, 3)
    >>> print(dist.fwd([-3, 0, 3])) # Forward Rosenblatt transformation
    [0.  0.5 1. ]

Here the two methods ``_cdf`` and ``_bnd`` are absolute requirements, while all
the others are nice to have. In addition to the once listed, it is also
possible to define the following methods:

``_mom``
    Method for creating raw statistical moments, used by the ``mom`` method.
``_ttr``
    Method for creating coefficients from three terms recursion method, used to
    perform "analytical" Stiltjes' method.
"""
import types
import numpy

from . import evaluation, approximation


class StochasticallyDependentError(Exception):
    """Error related to stochastically dependent variables."""


class Dist(object):
    """Baseclass for all probability distributions."""

    __array_priority__ = 9000
    """Numpy override variable."""
    _repr = None

    interpret_as_integer = False
    """
    Flag indicating that return value from the methods sample, range and inv
    should be interpreted as integers instead of floating point.
    """

    def __init__(self, **prm):
        """
        Args:
            prm (numpy.ndarray):
                Other optional parameters. Will be assumed when calling any
                sub-functions.
        """
        self.prm = prm


    def range(self, x_data=None):
        """
        Generate the upper and lower bounds of a distribution.

        Args:
            x_data (numpy.ndarray) :
                The bounds might vary over the sample space. By providing
                x_data you can specify where in the space the bound should be
                taken.  If omitted, a (pseudo-)random sample is used.

        Returns:
            (numpy.ndarray):
                The lower (out[0]) and upper (out[1]) bound where
                out.shape=(2,)+x_data.shape
        """
        if x_data is None:
            try:
                x_data = evaluation.evaluate_inverse(
                    self, numpy.array([[0.5]]*len(self)))
            except StochasticallyDependentError:
                x_data = approximation.find_interior_point(self)
            shape = (len(self),)
            if hasattr(self, "_range"):
                return self._range(x_data, {})
        else:
            x_data = numpy.asfarray(x_data)
            shape = x_data.shape
            x_data = x_data.reshape(len(self), -1)

        q_data = evaluation.evaluate_bound(self, x_data)
        q_data = q_data.reshape((2,)+shape)
        if self.interpret_as_integer:
            q_data = q_data.astype(int)
        return q_data

    def fwd(self, x_data):
        """
        Forward Rosenblatt transformation.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. ``x_data.shape`` must
                be compatible with distribution shape.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where
                ``out.shape==x_data.shape``.
        """
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)

        lower, upper = evaluation.evaluate_bound(self, x_data)
        q_data = numpy.zeros(x_data.shape)
        indices = x_data > upper
        q_data[indices] = 1
        indices = ~indices & (x_data >= lower)

        q_data[indices] = numpy.clip(evaluation.evaluate_forward(
            self, x_data), a_min=0, a_max=1)[indices]

        q_data = q_data.reshape(shape)
        return q_data

    def cdf(self, x_data):
        """
        Cumulative distribution function.

        Note that chaospy only supports cumulative distribution functions for
        stochastically independent distributions.

        Args:
            x_data (numpy.ndarray):
                Location for the distribution function. Assumes that
                ``len(x_data) == len(distribution)``.

        Returns:
            (numpy.ndarray):
                Evaluated distribution function values, where output has shape
                ``x_data.shape`` in one dimension and ``x_data.shape[1:]`` in
                higher dimensions.
        """
        if len(self) > 1 and evaluation.get_dependencies(*self):
            raise StochasticallyDependentError(
                "Cumulative distribution does not support dependencies.")
        q_data = self.fwd(x_data)
        if len(self) > 1:
            q_data = numpy.prod(q_data, 0)
        return q_data

    def inv(self, q_data, max_iterations=100, tollerance=1e-5):
        """
        Inverse Rosenblatt transformation.

        If possible the transformation is done analytically. If not possible,
        transformation is approximated using an algorithm that alternates
        between Newton-Raphson and binary search.

        Args:
            q_data (numpy.ndarray):
                Probabilities to be inverse. If any values are outside ``[0,
                1]``, error will be raised. ``q_data.shape`` must be compatible
                with distribution shape.
            max_iterations (int):
                If approximation is used, this sets the maximum number of
                allowed iterations in the Newton-Raphson algorithm.
            tollerance (float):
                If approximation is used, this set the error tolerance level
                required to define a sample as converged.

        Returns:
            (numpy.ndarray):
                Inverted probability values where
                ``out.shape == q_data.shape``.
        """
        q_data = numpy.asfarray(q_data)
        assert numpy.all((q_data >= 0) & (q_data <= 1)), "sanitize your inputs!"
        shape = q_data.shape
        q_data = q_data.reshape(len(self), -1)
        x_data = evaluation.evaluate_inverse(self, q_data)
        lower, upper = evaluation.evaluate_bound(self, x_data)
        x_data = numpy.clip(x_data, a_min=lower, a_max=upper)
        x_data = x_data.reshape(shape)
        if self.interpret_as_integer:
            x_data = x_data.astype(int)
        return x_data

    def pdf(self, x_data, step=1e-7):
        """
        Probability density function.

        If possible the density will be calculated analytically. If not
        possible, it will be approximated by approximating the one-dimensional
        derivative of the forward Rosenblatt transformation and multiplying the
        component parts. Note that even if the distribution is multivariate,
        each component of the Rosenblatt is one-dimensional.

        Args:
            x_data (numpy.ndarray):
                Location for the density function. ``x_data.shape`` must be
                compatible with distribution shape.
            step (float, numpy.ndarray):
                If approximation is used, the step length given in the
                approximation of the derivative. If array provided, elements
                are used along each axis.

        Returns:
            (numpy.ndarray):
                Evaluated density function values. Shapes are related through
                the identity ``x_data.shape == dist.shape+out.shape``.
        """
        x_data = numpy.asfarray(x_data)
        shape = x_data.shape
        x_data = x_data.reshape(len(self), -1)

        lower, upper = evaluation.evaluate_bound(self, x_data)
        f_data = numpy.zeros(x_data.shape)
        indices = (x_data <= upper) & (x_data >= lower)
        f_data[indices] = evaluation.evaluate_density(self, x_data)[indices]
        f_data = f_data.reshape(shape)
        if len(self) > 1:
            f_data = numpy.prod(f_data, 0)
        return f_data

    def sample(self, size=(), rule="R", antithetic=None):
        """
        Create pseudo-random generated samples.

        By default, the samples are created using standard (pseudo-)random
        samples. However, if needed, the samples can also be created by either
        low-discrepancy sequences, and/or variance reduction techniques.

        Changing the sampling scheme, use the following ``rule`` flag:

        +-------+-------------------------------------------------+
        | key   | Description                                     |
        +=======+=================================================+
        | ``C`` | Roots of the first order Chebyshev polynomials. |
        +-------+-------------------------------------------------+
        | ``NC``| Chebyshev nodes adjusted to ensure nested.      |
        +-------+-------------------------------------------------+
        | ``K`` | Korobov lattice.                                |
        +-------+-------------------------------------------------+
        | ``R`` | Classical (Pseudo-)Random samples.              |
        +-------+-------------------------------------------------+
        | ``RG``| Regular spaced grid.                            |
        +-------+-------------------------------------------------+
        | ``NG``| Nested regular spaced grid.                     |
        +-------+-------------------------------------------------+
        | ``L`` | Latin hypercube samples.                        |
        +-------+-------------------------------------------------+
        | ``S`` | Sobol low-discrepancy sequence.                 |
        +-------+-------------------------------------------------+
        | ``H`` | Halton low-discrepancy sequence.                |
        +-------+-------------------------------------------------+
        | ``M`` | Hammersley low-discrepancy sequence.            |
        +-------+-------------------------------------------------+

        All samples are created on the ``[0, 1]``-hypercube, which then is
        mapped into the domain of the distribution using the inverse Rosenblatt
        transformation.

        Args:
            size (numpy.ndarray):
                The size of the samples to generate.
            rule (str):
                Indicator defining the sampling scheme.
            antithetic (bool, numpy.ndarray):
                If provided, will be used to setup antithetic variables. If
                array, defines the axes to mirror.

        Returns:
            (numpy.ndarray):
                Random samples with shape ``(len(self),)+self.shape``.
        """
        size_ = numpy.prod(size, dtype=int)
        dim = len(self)
        if dim > 1:
            if isinstance(size, (tuple, list, numpy.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        from . import sampler
        out = sampler.generator.generate_samples(
            order=size_, domain=self, rule=rule, antithetic=antithetic)
        try:
            out = out.reshape(shape)
        except:
            if len(self) == 1:
                out = out.flatten()
            else:
                out = out.reshape(dim, int(out.size/dim))

        if self.interpret_as_integer:
            out = out.astype(int)
        return out

    def mom(self, K, **kws):
        """
        Raw statistical moments.

        Creates non-centralized raw moments from the random variable. If
        analytical options can not be utilized, Monte Carlo integration
        will be used.

        Args:
            K (numpy.ndarray):
                Index of the raw moments. k.shape must be compatible with
                distribution shape.  Sampling scheme when performing Monte
                Carlo
            rule (str):
                rule for estimating the moment if the analytical method fails.
            antithetic (numpy.ndarray):
                List of bool. Represents the axes to mirror using antithetic
                variable during MCI.

        Returns:
            (numpy.ndarray):
                Shapes are related through the identity
                ``k.shape == dist.shape+k.shape``.
        """
        K = numpy.asarray(K, dtype=int)
        shape = K.shape
        dim = len(self)

        if dim > 1:
            shape = shape[1:]

        size = int(K.size/dim)
        K = K.reshape(dim, size)

        cache = {}
        out = [evaluation.evaluate_moment(self, kdata, cache) for kdata in K.T]
        out = numpy.array(out)
        return out.reshape(shape)

    def _mom(self, *args, **kws):
        """Default moment generator, throws error."""
        raise StochasticallyDependentError("component lack support")

    def ttr(self, kloc, acc=10**3, verbose=1):
        """
        Three terms relation's coefficient generator

        Args:
            k (numpy.ndarray, int):
                The order of the coefficients.
            acc (int):
                Accuracy of discretized Stieltjes if analytical methods are
                unavailable.

        Returns:
            (Recurrence coefficients):
                Where out[0] is the first (A) and out[1] is the second
                coefficient With ``out.shape==(2,)+k.shape``.
        """
        kloc = numpy.asarray(kloc, dtype=int)
        shape = kloc.shape
        kloc = kloc.reshape(len(self), -1)
        cache = {}
        out = [evaluation.evaluate_recurrence_coefficients(self, k)
               for k in kloc.T]
        alpha, beta = numpy.asfarray(list(zip(*out)))
        out = numpy.array([alpha.T, beta.T])
        return out.reshape((2,)+shape)


    def _ttr(self, kloc, cache, **kws):
        """Default TTR generator, throws error."""
        raise NotImplementedError()

    def __str__(self):
        """X.__str__() <==> str(X)"""
        if self._repr is not None:
            kwargs = self._repr
        else:
            kwargs = self.prm
        args = [key + "=" + str(kwargs[key]) for key in sorted(kwargs)]
        return self.__class__.__name__ + "(" + ", ".join(args) + ")"

    def __repr__(self):
        return str(self)

    def __len__(self):
        """X.__len__() <==> len(X)"""
        return 1

    def __add__(self, X):
        """Y.__add__(X) <==> X+Y"""
        from . import operators
        return operators.Add(self, X)

    def __radd__(self, X):
        """Y.__radd__(X) <==> Y+X"""
        from . import operators
        return operators.Add(self, X)

    def __sub__(self, X):
        """Y.__sub__(X) <==> X-Y"""
        from . import operators
        return operators.Add(self, -X)

    def __rsub__(self, X):
        """Y.__rsub__(X) <==> Y-X"""
        from . import operators
        return operators.Add(X, -self)

    def __neg__(self):
        """X.__neg__() <==> -X"""
        from . import operators
        return operators.Neg(self)

    def __mul__(self, X):
        """Y.__mul__(X) <==> X*Y"""
        from . import operators
        return operators.Mul(self, X)

    def __rmul__(self, X):
        """Y.__rmul__(X) <==> Y*X"""
        from . import operators
        return operators.Mul(X, self)

    def __div__(self, X):
        """Y.__div__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rdiv__(self, X):
        """Y.__rdiv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __floordiv__(self, X):
        """Y.__floordiv__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rfloordiv__(self, X):
        """Y.__rfloordiv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __truediv__(self, X):
        """Y.__truediv__(X) <==> Y/X"""
        from . import operators
        return operators.Mul(self, X**-1)

    def __rtruediv__(self, X):
        """Y.__rtruediv__(X) <==> X/Y"""
        from . import operators
        return operators.Mul(X, self**-1)

    def __pow__(self, X):
        """Y.__pow__(X) <==> Y**X"""
        from . import operators
        return operators.Pow(self, X)

    def __rpow__(self, X):
        """Y.__rpow__(X) <==> X**Y"""
        from . import operators
        return operators.Pow(X, self)

    def __le__(self, X):
        """Y.__le__(X) <==> Y<=X"""
        from . import operators
        return operators.Trunc(self, X)

    def __lt__(self, X):
        """Y.__lt__(X) <==> Y<X"""
        from . import operators
        return operators.Trunc(self, X)

    def __ge__(self, X):
        """Y.__ge__(X) <==> Y>=X"""
        from . import operators
        return operators.Trunc(X, self)

    def __gt__(self, X):
        """Y.__gt__(X) <==> Y>X"""
        from . import operators
        return operators.Trunc(X, self)



"""
Custom distribution constructor.

Example usage
-------------

Construct distribution from scratch::

    >>> MyUniform = chaospy.construct(
    ...     cdf=lambda self, x, lo, up: (x - lo) / (up - lo),
    ...     bnd=lambda self, x, lo, up: (lo, up)
    ... )

Evaluate distribution::

    >>> uniform = MyUniform(lo=-1, up=1)
    >>> print(uniform.pdf(numpy.linspace(-2, 2, 16)))
    [0.  0.  0.  0.  0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.  0.  0.  0. ]
"""
import types
from .baseclass import Dist

LEGAL_ATTRS = {
    "cdf": "_cdf", "bnd": "_bnd",
    "pdf": "_pdf", "ppf": "_ppf", "mom": "_mom",
    "ttr": "_ttr", "fwd_cache": "_fwd_cache", "inv_cache": "_inv_cache",
    "doc": "__doc__",
    "str": "_str"
}

def construct(parent=None, defaults=None, **kwargs):
    """
    Random variable constructor.

    Args:
        cdf:
            Cumulative distribution function. Optional if ``parent`` is used.
        bnd:
            Boundary interval. Optional if ``parent`` is used.
        parent (Dist):
            Distribution used as basis for new distribution. Any other argument
            that is omitted will instead take is function from ``parent``.
        doc (str]):
            Documentation for the distribution.
        str (str, :py:data:typing.Callable):
            Pretty print of the variable.
        pdf:
            Probability density function.
        ppf:
            Point percentile function.
        mom:
            Raw moment generator.
        ttr:
            Three terms recursion coefficient generator.
        init:
            Custom initialiser method.
        defaults (dict):
            Default values to provide to initialiser.

    Returns:
        (Dist):
            New custom distribution.
    """
    for key in kwargs:
        assert key in LEGAL_ATTRS, "{} is not legal input".format(key)

    if parent is not None:
        for key, value in LEGAL_ATTRS.items():
            if key not in kwargs and hasattr(parent, value):
                    kwargs[key] = getattr(parent, value)

    assert "cdf" in kwargs, "cdf function must be defined"
    assert "bnd" in kwargs, "bnd function must be defined"
    if "str" in kwargs and isinstance(kwargs["str"], str):
        string = kwargs.pop("str")
        kwargs["str"] = lambda *args, **kwargs: string

    defaults = defaults if defaults else {}
    for key in defaults:
        assert key in LEGAL_ATTRS, "invalid default value {}".format(key)

    def custom_distribution(**kws):

        prm = defaults.copy()
        prm.update(kws)
        dist = Dist(**prm)

        for key, function in kwargs.items():
            attr_name = LEGAL_ATTRS[key]
            setattr(dist, attr_name, types.MethodType(function, dist))
        return dist

    if "doc" in kwargs:
        custom_distribution.__doc__ = kwargs["doc"]

    return custom_distribution



"""Conditional expected value."""
from itertools import product
import numpy

from .. import distributions, poly as polynomials, quadrature


def E_cond(poly, freeze, dist, **kws):
    """
    Conditional expected value operator.

    1st order statistics of a polynomial on a given probability space
    conditioned on some of the variables.

    Args:
        poly (Poly):
            Polynomial to find conditional expected value on.
        freeze (numpy.ndarray):
            Boolean values defining the conditional variables. True values
            implies that the value is conditioned on, e.g. frozen during the
            expected value calculation.
        dist (Dist) :
            The distributions of the input used in ``poly``.

    Returns:
        (chaospy.poly.base.Poly) :
            Same as ``poly``, but with the variables not tagged in ``frozen``
            integrated away.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> print(chaospy.E_cond(poly, [1, 0], dist))
        [1.0, q0, 0.0, 0.0]
        >>> print(chaospy.E_cond(poly, [0, 1], dist))
        [1.0, 1.0, q1, 10.0q1]
        >>> print(chaospy.E_cond(poly, [1, 1], dist))
        [1.0, q0, q1, 10.0q0q1]
        >>> print(chaospy.E_cond(poly, [0, 0], dist))
        [1.0, 1.0, 0.0, 0.0]
    """
    if poly.dim < len(dist):
        poly = polynomials.setdim(poly, len(dist))

    freeze = polynomials.Poly(freeze)
    freeze = polynomials.setdim(freeze, len(dist))
    keys = freeze.keys
    if len(keys) == 1 and keys[0] == (0,)*len(dist):
        freeze = list(freeze.A.values())[0]
    else:
        freeze = numpy.array(keys)
    freeze = freeze.reshape(int(freeze.size/len(dist)), len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    kmax = numpy.max(poly.keys, 0) + 1
    keys = [range(k) for k in kmax]

    A = poly.A.copy()
    keys = poly.keys
    out = {}
    zeros = [0]*poly.dim

    for i in range(len(keys)):

        key = list(keys[i])
        a = A[tuple(key)]

        for d in range(poly.dim):
            for j in range(len(freeze)):
                if freeze[j, d]:
                    key[d], zeros[d] = zeros[d], key[d]
                    break

        tmp = a*dist.mom(tuple(key))
        if tuple(zeros) in out:
            out[tuple(zeros)] = out[tuple(zeros)] + tmp
        else:
            out[tuple(zeros)] = tmp

        for d in range(poly.dim):
            for j in range(len(freeze)):
                if freeze[j, d]:
                    key[d], zeros[d] = zeros[d], key[d]
                    break

    out = polynomials.Poly(out, poly.dim, poly.shape, float)
    out = polynomials.reshape(out, shape)

    return out



"""
Historically, the syntax for ``chaospy``'s distributions is inspired by the
``scipy.stats`` distributions using the syntax. Over time though, they have
developed into different entities making them incompatible. The
:class:`ScipyStatsDist` exists to restore this incompatibility again.

Basic usage of the :class:`ScipyStatsDist` constructor involves just passing
the ``scipy.stats`` distributions as input argument::

    >>> from scipy.stats import norm
    >>> st_distribution = norm(0, 1)
    >>> distribution = chaospy.ScipyStatsDist(st_distribution)
    >>> print(distribution)
    ScipyStatsDist(norm(0, 1))

This distribution then behaves as a normal ``chaospy`` distribution::

    >>> print(numpy.around(distribution.pdf([-1, 0, 1]), 4))
    [0.242  0.3989 0.242 ]
    >>> print(distribution.mom([0, 1, 2]))
    [1. 0. 1.]

Currently the wrapper is limited to only support univariate distributions.
"""
import numpy

from ..distributions import Dist


class ScipyStatsDist(Dist):
    """
    One dimensional ``scipy.stats`` distribution.

    Args:
        distribution (openturns.Distribution):
            1D distribution created with ``scipy.stats``.
    """

    def __init__(self, distribution):
        from scipy.stats._distn_infrastructure import rv_frozen
        assert isinstance(distribution, rv_frozen), (
            "Expected frozen distribution from ``scipy.stats``.")
        assert not hasattr(distribution, "dim") or distribution.dim == 1, (
            "Only one-dimensional ``scipy.stats`` models supported.")
        Dist.__init__(self)
        self.distribution = distribution

    def _pdf(self, x_loc):
        return self.distribution.pdf(x_loc)

    def _cdf(self, x_loc):
        return self.distribution.cdf(x_loc)

    def _ppf(self, q_loc):
        return self.distribution.ppf(q_loc)

    def _bnd(self, x_loc):
        return self.distribution.interval(1-1e-14)

    def _mom(self, k_loc):
        return self.distribution.moment(int(k_loc))

    def __str__(self):
        name = self.distribution.dist.__class__.__name__
        name = name.replace("_gen", "")
        args = ", ".join([str(arg) for arg in self.distribution.args])
        return "%s(%s(%s))" % (self.__class__.__name__, name, args)



"""Arc-Cosines Hyperbolic."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Arccosh(Dist):
    """
    Arc-Cosine Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Arccosh(chaospy.Uniform(1, 2))
        >>> print(distribution)
        Arccosh(Uniform(lower=1, upper=2))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.6224 0.867  1.047  1.1929]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6633 0.9798 1.249  1.4967]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.0887 0.4751 1.2878 0.9463]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.9019
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.9019 0.7012 0.6772]
         [1.     0.0933 0.1037]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() >= 1)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        return evaluation.evaluate_density(
            dist, numpy.cosh(x), cache=cache)*numpy.sinh(x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.cosh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.arccosh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.arccosh(evaluation.evaluate_bound(
            dist, numpy.cosh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arccosh(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.cosh(dist)
        return self



"""Double gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class double_gamma(Dist):
    """Double gamma distribution."""

    def __init__(self, a):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        ax = abs(x)
        return 1.0/(2*special.gamma(a))*ax**(a-1.0) * numpy.exp(-ax)

    def _cdf(self, x, a):
        fac = 0.5*special.gammainc(a,abs(x))
        return numpy.where(x>0,0.5+fac,0.5-fac)

    def _ppf(self, q, a):
        fac = special.gammainccinv(a,1-abs(2*q-1))
        return numpy.where(q>0.5, fac, -fac)

    def _bnd(self, x, a):
        return self._ppf(1e-10, a), self._ppf(1-1e-10, a)


class DoubleGamma(Add):
    """
    Double gamma distribution.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.DoubleGamma(2, 4, 2)
        >>> print(distribution)
        DoubleGamma(scale=4, shape=2, shift=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [-100.4566   -4.7134    2.        8.7134  104.4566]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.0392 0.     0.0392 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 6.4679 -9.2251 17.5874  0.8239]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=double_gamma(shape)*scale, right=shift)


Dbl_gamma = deprecation_warning(DoubleGamma, "Dbl_gamma")



"""
Evaluate distribution bounds: The location where it is reasonable to define the
distribution density is located between.

Example usage
-------------

Define a simple distribution and data::

    >>> class Uniform(chaospy.Dist):
    ...     def _bnd(self, x_data, lo, up): return lo, up
    >>> dist = Uniform(lo=1, up=3)
    >>> x_data = numpy.array([[0.1, 0.2, 0.3]])

Normal usage::

    >>> lower, upper = evaluate_bound(dist, x_data)
    >>> print(numpy.around(lower, 4))
    [[1. 1. 1.]]
    >>> print(numpy.around(upper, 4))
    [[3. 3. 3.]]
    >>> lower, upper = evaluate_bound(dist, x_data, parameters={"lo": -1})
    >>> print(numpy.around(lower, 4))
    [[-1. -1. -1.]]
    >>> print(numpy.around(upper, 4))
    [[3. 3. 3.]]
"""
import numpy
from .parameters import load_parameters


def evaluate_bound(
        distribution,
        x_data,
        parameters=None,
        cache=None,
):
    """
    Evaluate lower and upper bounds.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        x_data (numpy.ndarray):
            Locations for where evaluate bounds at. Relevant in the case of
            multivariate distributions where the bounds are affected by the
            output of other distributions.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The lower and upper bounds of ``distribution`` at location
        ``x_data`` using parameters ``parameters``.
    """
    assert len(x_data) == len(distribution)
    assert len(x_data.shape) == 2

    cache = cache if cache is not None else {}

    parameters = load_parameters(
        distribution, "_bnd", parameters=parameters, cache=cache)

    out = numpy.zeros((2,) + x_data.shape)

    lower, upper = distribution._bnd(x_data.copy(), **parameters)
    out.T[:, :, 0] = numpy.asfarray(lower).T
    out.T[:, :, 1] = numpy.asfarray(upper).T

    cache[distribution] = out

    return out



"""T-Copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist

class t_copula(Dist):
    """T-Copula."""

    def __init__(self, df, R):
        C = numpy.linalg.cholesky(R)
        Ci = numpy.linalg.inv(C)
        self.length = len(R)
        Dist.__init__(self, df=df, C=C, Ci=Ci)

    def __len__(self):
        return self.length

    def _cdf(self, x, df, C, Ci):
        out = special.stdtr(df, numpy.dot(Ci, special.stdtrit(df, x)))
        return out

    def _ppf(self, q, df, C, Ci):
        out = special.stdtr(df, numpy.dot(C, special.stdtrit(df, q)))
        return out

    def _bnd(self):
        return 0., 1.


class TCopula(Copula):
    """
    T-Copula.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Uniform(), 2)
        >>> R = [[1, 0.5], [0.5, 1]]
        >>> copula = chaospy.TCopula(distribution, 5, R)
        >>> print(copula)
        TCopula(Iid(Uniform(lower=0, upper=1), 2), R=[[1, 0.5], [0.5, 1]], df=5)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(copula.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.1832 0.2784 0.4004]
          [0.3656 0.5    0.6344]
          [0.5996 0.7216 0.8168]]]
        >>> print(numpy.around(copula.fwd(copula.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(copula.pdf(copula.inv(mesh)), 4))
        [[1.4656 1.0739 0.8912]
         [1.2486 1.1547 1.2486]
         [0.8912 1.0739 1.4656]]
        >>> print(numpy.around(copula.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.8783 0.1053 0.2107 0.4021]]
    """

    def __init__(self, dist, df, R):
        """
        Args:
            dist (Dist):
                The Distribution to wrap in a copula.
            R (numpy.ndarray):
                Covariance matrix defining dependencies..
            df (float):
                The degree of freedom in the underlying student-t distribution.
        """
        self._repr = {"df": df, "R": R}
        Copula.__init__(self, dist=dist, trans=t_copula(df, R))



"""
Clenshaw-Curtis quadrature method is a good all-around quadrature method
comparable to Gaussian quadrature, but typically limited to finite intervals
without a specific weight function. In addition to be quite accurate, the
weights and abscissas can be calculated quite fast.

Another thing to note is that Clenshaw-Curtis, with an appropriate growth rule
is fully nested. This means, if one applies a method that combines different
order of quadrature rules, the number of evaluations can often be reduced as
the abscissas can be used across levels.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis")
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0. 1.]] [0.5 0.5]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.   0.25 0.75 1.  ]] [0.056 0.444 0.444 0.056]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="clenshaw_curtis", growth=True)
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    2 [[0.    0.146 0.5   0.854 1.   ]] [0.033 0.267 0.4   0.267 0.033]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> X, W = chaospy.generate_quadrature(
    ...     2, distribution, rule="clenshaw_curtis",
    ...     growth=True, sparse=True)
    >>> print(numpy.around(X, 2))
    [[0.   0.   0.   0.15 0.5  0.5  0.5  0.5  0.5  0.85 1.   1.   1.  ]
     [0.   0.5  1.   0.5  0.   0.15 0.5  0.85 1.   0.5  0.   0.5  1.  ]]
    >>> print(numpy.around(W, 3))
    [ 0.028 -0.022  0.028  0.267 -0.022  0.267 -0.089  0.267 -0.022  0.267
      0.028 -0.022  0.028]
"""
from __future__ import division, print_function

import numpy

from .combine import combine_quadrature


def quad_clenshaw_curtis(order, domain, growth=False):
    """
    Generate the quadrature nodes and weights in Clenshaw-Curtis quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Dist, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_clenshaw_curtis(3, (0, 1))
        >>> print(numpy.around(abscissas, 4))
        [[0.   0.25 0.75 1.  ]]
        >>> print(numpy.around(weights, 4))
        [0.0556 0.4444 0.4444 0.0556]
    """
    from ..distributions.baseclass import Dist
    if isinstance(domain, Dist):
        abscissas, weights = quad_clenshaw_curtis(
            order, domain.range(), growth)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if growth:
        order = numpy.where(order > 0, 2**order, 0)

    abscissas, weights = zip(*[_clenshaw_curtis(order_) for order_ in order])

    return combine_quadrature(abscissas, weights, (lower, upper))


def _clenshaw_curtis(order):
    r"""
    Backend method.

    Examples:
        >>> abscissas, weights = _clenshaw_curtis(0)
        >>> print(abscissas)
        [0.5]
        >>> print(weights)
        [1.]
        >>> abscissas, weights = _clenshaw_curtis(1)
        >>> print(abscissas)
        [0. 1.]
        >>> print(weights)
        [0.5 0.5]
        >>> abscissas, weights = _clenshaw_curtis(2)
        >>> print(abscissas)
        [0.  0.5 1. ]
        >>> print(weights)
        [0.16666667 0.66666667 0.16666667]
        >>> abscissas, weights = _clenshaw_curtis(3)
        >>> print(abscissas)
        [0.   0.25 0.75 1.  ]
        >>> print(weights)
        [0.05555556 0.44444444 0.44444444 0.05555556]
        >>> abscissas, weights = _clenshaw_curtis(4)
        >>> print(abscissas)
        [0.         0.14644661 0.5        0.85355339 1.        ]
        >>> print(weights)
        [0.03333333 0.26666667 0.4        0.26666667 0.03333333]
        >>> abscissas, weights = _clenshaw_curtis(5)
        >>> print(abscissas)
        [0.        0.0954915 0.3454915 0.6545085 0.9045085 1.       ]
        >>> print(weights)
        [0.02       0.18037152 0.29962848 0.29962848 0.18037152 0.02      ]
    """
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscisas = 0.5*numpy.cos(theta) + 0.5

    idx, idy = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(idy+1)*theta[idx])/(4*idy*(idy+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    weights[0] /= 2
    weights[-1] /= 2

    return abscisas, weights



r"""
Copulas are a type dependency structure imposed on independent variables to
achieve to more complex problems without adding too much complexity.

To construct a copula one needs a copula transformation and the
Copula wrapper::

    >>> dist = chaospy.Iid(chaospy.Uniform(), 2)
    >>> copula = chaospy.Gumbel(dist, theta=1.5)

The resulting copula is then ready for use::

    >>> print(numpy.around(copula.sample(5), 4))
    [[0.6536 0.115  0.9503 0.4822 0.8725]
     [0.2483 0.3325 0.1725 0.3206 0.2732]]
"""
from .baseclass import Copula, Archimedean

from .gumbel import Gumbel
from .clayton import Clayton
from .ali_mikhail_haq import AliMikhailHaq
from .frank import Frank
from .joe import Joe
from .nataf import Nataf
from .t_copula import TCopula



"""Common functions and attributes used across submodules."""
import inspect


class DependencyError(ValueError):
    """Error that occurs with bad stochastic dependency structures."""


def contains_call_signature(caller, key):
    """
    Check if a function or method call signature contains a specific
    argument.

    Args:
        caller (Callable):
            Method or function to check if signature is contain in.
        key (str):
            Signature to look for.

    Returns:
        True if ``key`` exits in ``caller`` call signature.

    Examples:
        >>> def foo(param): pass
        >>> contains_call_signature(foo, "param")
        True
        >>> contains_call_signature(foo, "not_param")
        False
        >>> class Bar:
        ...     def baz(self, param): pass
        >>> bar = Bar()
        >>> contains_call_signature(bar.baz, "param")
        True
        >>> contains_call_signature(bar.baz, "not_param")
        False
    """
    try:
        args = inspect.signature(caller).parameters
    except AttributeError:
        args = inspect.getargspec(caller).args
    return key in args



"""Burr Type III distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class burr(Dist):
    """Stadard Burr distribution."""

    def __init__(self, alpha=1., kappa=1.):
        Dist.__init__(self, alpha=alpha, kappa=kappa)

    def _pdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = alpha*kappa*x[indices]**(-alpha-1.0)
        output[indices] /= (1+x[indices]**alpha)**(kappa+1)
        return output

    def _cdf(self, x, alpha, kappa):
        output = numpy.zeros(x.shape)
        indices = x > 0
        output[indices] = 1-(1+x[indices]**alpha)**-kappa
        return output

    def _ppf(self, q, alpha, kappa):
        return ((1-q)**(-1./kappa) -1)**(1./alpha)

    def _mom(self, k, alpha, kappa):
        return kappa*special.beta(1-k*1./alpha, kappa+k*1./alpha)

    def _bnd(self, x, alpha, kappa):
        return 0, self._ppf(1-1e-10, alpha, kappa)


class Burr(Add):
    """
    Burr Type XII or Singh-Maddala distribution.

    Args:
        alpha (float, Dist): First shape parameter
        kappa (float, Dist): Second shape parameter
        loc (float, Dist): Location parameter
        scale (float, Dist): Scaling parameter

    Examples:
        >>> distribution = chaospy.Burr(100, 1.2, 4, 2)
        >>> print(distribution)
        Burr(alpha=100, kappa=1.2, loc=4, scale=2)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [5.9642 5.9819 5.9951 6.0081 6.0249]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [266.5437  71.6255  21.5893   5.3229   0.643 ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [6.007  5.9558 6.0489 5.9937]
        >>> print(numpy.around(distribution.mom(1), 4))
        6.0061
    """

    def __init__(self, alpha=1, kappa=1, loc=0, scale=1):
        self._repr = {
            "alpha": alpha, "kappa": kappa, "loc": loc, "scale": scale}
        Add.__init__(self, left=burr(alpha, kappa)*scale, right=loc)



"""Modified Chebyshev algorithm."""
from __future__ import division

import numpy


def modified_chebyshev(moments):
    r"""
    Given the first 2N raw statistical moments, this method uses the modified
    Chebyshev algorithm for computing the associated recurrence coefficients.

    Args:
        moments (numpy.ndarray):
            Raw statistical moments from calculating the integrals
            :math:`\int x^k p(x) dx` for :math:`k=0,\dots,2N`.

    Examples:
        >>> dist = chaospy.Normal()
        >>> alpha, beta = modified_chebyshev(dist.mom(numpy.arange(8)))
        >>> print(alpha)
        [0. 0. 0. 0.]
        >>> print(beta)
        [1. 1. 2. 3.]
        >>> dist = chaospy.Uniform(-1, 1)
        >>> alpha, beta = modified_chebyshev(dist.mom(numpy.arange(8)))
        >>> print(alpha)
        [0. 0. 0. 0.]
        >>> print(beta)
        [1.         0.33333333 0.26666667 0.25714286]
    """
    moments = numpy.asfarray(moments).flatten()
    order = len(moments)
    assert order%2 == 0

    sigma = numpy.zeros((3, order))
    sigma[0] = moments
    coeffs = [(sigma[0, 1]/sigma[0, 0], sigma[0, 0])]

    for idx in range(1, order//2):
        sigma[idx%3, idx:order-idx] = (
            sigma[(idx-1)%3, idx+1:order-idx+1]-
            coeffs[idx-1][0]*sigma[(idx-1)%3, idx:order-idx]-
            coeffs[idx-1][1]*sigma[(idx-2)%3, idx:order-idx]
        )
        coeffs.append((
            (sigma[idx%3, idx+1]/sigma[idx%3, idx]
             -sigma[(idx-1)%3, idx]/sigma[(idx-1)%3, idx-1]),
            sigma[idx%3, idx]/sigma[(idx-1)%3, idx-1],
        ))

    coeffs = numpy.array(coeffs[:order//2]).reshape(-1, 2).T
    return coeffs



"""
Multiplication of distributions.

Example usage
-------------

Distribution * a constant::

    >>> distribution = chaospy.Uniform(0, 1) * 4
    >>> print(distribution)
    Mul(Uniform(lower=0, upper=1), 4)
    >>> print(numpy.around(distribution.sample(5), 4))
    [2.6144 0.46   3.8011 1.9288 3.4899]
    >>> print(numpy.around(distribution.fwd([1, 2, 3]), 4))
    [0.25 0.5  0.75]
    >>> print(numpy.around(distribution.inv(distribution.fwd([1, 2, 3])), 4))
    [1. 2. 3.]
    >>> print(numpy.around(distribution.pdf([1, 2, 3]), 4))
    [0.25 0.25 0.25]
    >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
    [ 2.      5.3333 16.    ]
    >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    [[2.     2.     2.    ]
     [1.3333 1.0667 1.0286]]

Construct joint multiplication distribution::

    >>> lhs = chaospy.Uniform(-1, 0)
    >>> rhs = chaospy.Uniform(-3, -2)
    >>> multiplication = lhs * rhs
    >>> print(multiplication)
    Mul(Uniform(lower=-1, upper=0), Uniform(lower=-3, upper=-2))
    >>> joint1 = chaospy.J(lhs, multiplication)
    >>> print(joint1.range())
    [[-1.  0.]
     [ 0.  3.]]
    >>> joint2 = chaospy.J(rhs, multiplication)
    >>> print(joint2.range())
    [[-3. -0.]
     [-2.  3.]]
    >>> joint3 = chaospy.J(multiplication, lhs)
    >>> print(joint3.range())
    [[ 0. -1.]
     [ 3.  0.]]
    >>> joint4 = chaospy.J(multiplication, rhs)
    >>> print(joint4.range())
    [[-0. -3.]
     [ 3. -2.]]

Generate random samples::

    >>> print(numpy.around(joint1.sample(4), 4))
    [[-0.7877 -0.9593 -0.6028 -0.7669]
     [ 2.2383  2.1172  1.6532  1.8345]]
    >>> print(numpy.around(joint2.sample(4), 4))
    [[-2.8177 -2.2565 -2.9304 -2.1147]
     [ 2.6843  2.1011  1.2174  0.0613]]

Forward transformations::

    >>> lcorr = numpy.array([-0.9, -0.5, -0.1])
    >>> rcorr = numpy.array([-2.99, -2.5, -2.01])
    >>> print(numpy.around(joint1.fwd([lcorr, lcorr*rcorr]), 4))
    [[0.1  0.5  0.9 ]
     [0.99 0.5  0.01]]
    >>> print(numpy.around(joint2.fwd([rcorr, lcorr*rcorr]), 4))
    [[0.01 0.5  0.99]
     [0.9  0.5  0.1 ]]

Inverse transformations::

    >>> print(numpy.around(joint1.inv(joint1.fwd([lcorr, lcorr*rcorr])), 4))
    [[-0.9   -0.5   -0.1  ]
     [ 2.691  1.25   0.201]]
    >>> print(numpy.around(joint2.inv(joint2.fwd([rcorr, lcorr*rcorr])), 4))
    [[-2.99  -2.5   -2.01 ]
     [ 2.691  1.25   0.201]]
"""
import numpy

from ..baseclass import Dist
from .. import evaluation


class Mul(Dist):
    """Multiplication."""

    def __init__(self, left, right):
        """
        Args:
            left (Dist, numpy.ndarray) : Left hand side.
            right (Dist, numpy.ndarray) : Right hand side.
        """
        self.matrix = False
        if (isinstance(left, Dist) and
                len(left) > 1 and
                not isinstance(right, Dist)):
            right = right*numpy.eye(len(left))
            self.matrix = True

        elif (isinstance(right, Dist) and
                len(right) > 1 and
                not isinstance(left, Dist)):
            left = left*numpy.eye(len(right))
            self.matrix = True

        Dist.__init__(self, left=left, right=right)


    def _bnd(self, xloc, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> print(chaospy.Uniform().range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [1. 1. 1. 1.]]
            >>> print(Mul(chaospy.Uniform(), 2).range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [2. 2. 2. 2.]]
            >>> print(Mul(2, chaospy.Uniform()).range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [2. 2. 2. 2.]]
            >>> print(Mul(2, 2).range([-2, 0, 2, 4]))
            [[4. 4. 4. 4.]
             [4. 4. 4. 4.]]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.range([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [[[0. 0. 0.]
              [0. 0. 0.]]
            <BLANKLINE>
             [[1. 1. 1.]
              [2. 2. 2.]]]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.range([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [[[0. 0. 0.]
              [0. 0. 0.]]
            <BLANKLINE>
             [[2. 2. 2.]
              [1. 1. 1.]]]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.range([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [[[0. 0. 0.]
              [0. 0. 0.]]
            <BLANKLINE>
             [[1. 1. 1.]
              [2. 2. 2.]]]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return numpy.dot(left, right), numpy.dot(left, right)
        else:
            left = numpy.asfarray(left)
            if self.matrix:
                Ci = numpy.linalg.inv(left)
                xloc = numpy.dot(Ci, xloc)
                assert len(xloc) == len(right)

            elif len(left.shape) == 3:
                left_ = numpy.mean(left, 0)
                valids = left_ != 0
                xloc.T[valids.T] = xloc.T[valids.T]/left_.T[valids.T]

            else:
                left = (left.T+numpy.zeros(xloc.shape).T).T
                valids = left != 0
                xloc.T[valids.T] = xloc.T[valids.T]/left.T[valids.T]

            assert len(xloc) == len(right)
            lower, upper = evaluation.evaluate_bound(right, xloc, cache=cache)
            if self.matrix:
                lower = numpy.dot(lower.T, left.T).T
                upper = numpy.dot(upper.T, left.T).T

            elif len(left.shape) == 3:
                lower = numpy.where(left[0]*lower > 0, left[0]*lower, left[1]*lower)
                upper = numpy.where(left[1]*upper > 0, left[1]*upper, left[0]*upper)
                lower, upper = (
                    numpy.where(lower < upper, lower, upper),
                    numpy.where(lower < upper, upper, lower),
                )
                lower[(left[0] < 0) & (lower > 0)] = 0.
                assert len(lower) == len(right)

            else:
                lower *= left
                upper *= left
            lower, upper = (
                numpy.where(lower < upper, lower, upper),
                numpy.where(lower < upper, upper, lower),
            )
            return lower, upper


        right = numpy.asfarray(right)
        if self.matrix:
            Ci = numpy.linalg.inv(right)
            xloc = numpy.dot(xloc.T, Ci.T).T
            assert len(left) == len(xloc)

        elif len(right.shape) == 3:
            right_ = numpy.mean(right, 0)
            valids = right_ != 0
            xloc.T[valids.T] = xloc.T[valids.T]/right_.T[valids.T]

        else:
            right = (right.T+numpy.zeros(xloc.shape).T).T
            valids = right != 0
            xloc.T[valids.T] = xloc.T[valids.T]/right.T[valids.T]

        assert len(left) == len(xloc)
        lower, upper = evaluation.evaluate_bound(left, xloc, cache=cache)
        if self.matrix:
            lower = numpy.dot(lower.T, right.T).T
            upper = numpy.dot(upper.T, right.T).T

        elif len(right.shape) == 3:
            lower = numpy.where(right[0]*lower > 0, right[0]*lower, right[1]*lower)
            upper = numpy.where(right[1]*upper > 0, right[1]*upper, right[0]*upper)

            lower, upper = (
                numpy.where(lower < upper, lower, upper),
                numpy.where(lower < upper, upper, lower),
            )
            lower[(right[0] < 0) & (lower > 0)] = 0.

        else:
            lower *= right
            upper *= right
        lower, upper = (
            numpy.where(lower < upper, lower, upper),
            numpy.where(lower < upper, upper, lower),
        )
        assert lower.shape == xloc.shape
        assert upper.shape == xloc.shape
        return lower, upper

    def _cdf(self, xloc, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> print(chaospy.Uniform().fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 1.  1. ]
            >>> print(Mul(chaospy.Uniform(), 2).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.   0.25 0.75 1.  ]
            >>> print(Mul(2, chaospy.Uniform()).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.   0.25 0.75 1.  ]
            >>> print(Mul(1, 1.5).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.  0.5 1. ]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.fwd([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [[0.25 0.3  0.75]
             [0.5  0.6  1.  ]]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.fwd([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [[0.5  0.6  1.  ]
             [0.25 0.3  0.75]]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            if self.matrix:
                return 0.5*(numpy.dot(left, right) == xloc)
            return 0.5*(left*right == xloc)
        else:
            if self.matrix:
                Ci = numpy.linalg.inv(left)
                xloc = numpy.dot(Ci, xloc)
                assert len(xloc) == len(numpy.dot(Ci, xloc))

            else:
                left = (numpy.asfarray(left).T+numpy.zeros(xloc.shape).T).T
                valids = left != 0
                xloc.T[valids.T] = xloc.T[valids.T]/left.T[valids.T]

            uloc = evaluation.evaluate_forward(right, xloc, cache=cache)
            if not self.matrix:
                uloc = numpy.where(left.T >= 0, uloc.T, 1-uloc.T).T
            assert uloc.shape == xloc.shape
            return uloc

        if self.matrix:
            Ci = numpy.linalg.inv(right)
            xloc = numpy.dot(xloc.T, Ci).T
        else:
            right = (numpy.asfarray(right).T+numpy.zeros(xloc.shape).T).T
            valids = right != 0
            xloc.T[valids.T] = xloc.T[valids.T]/right.T[valids.T]

        assert len(left) == len(xloc)
        uloc = evaluation.evaluate_forward(left, xloc, cache=cache)
        if not self.matrix:
            uloc = numpy.where(right.T >= 0, uloc.T, 1-uloc.T).T
        assert uloc.shape == xloc.shape
        return uloc

    def _ppf(self, uloc, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(Mul(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9]))
            [0.2 0.4 1.8]
            >>> print(Mul(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [0.2 0.4 1.8]
            >>> print(Mul(2, 2).inv([0.1, 0.2, 0.9]))
            [4. 4. 4.]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]]))
            [[1.  1.2 1.4]
             [0.5 0.6 0.7]]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.inv([[0.5, 0.6, 0.7], [0.5, 0.6, 0.7]]))
            [[0.5 0.6 0.7]
             [1.  1.2 1.4]]
        """
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            if self.matrix:
                return numpy.dot(left, right)
            return left*right

        else:
            if not self.matrix:
                uloc = numpy.where(numpy.asfarray(left).T > 0, uloc.T, 1-uloc.T).T
            xloc = evaluation.evaluate_inverse(right, uloc, cache=cache)
            if self.matrix:
                xloc = numpy.dot(left, xloc)
            else:
                xloc *= left
            return xloc

        if not self.matrix:
            uloc = numpy.where(numpy.asfarray(right).T > 0, uloc.T, 1-uloc.T).T
        xloc = evaluation.evaluate_inverse(left, uloc, cache=cache)
        if self.matrix:
            xloc = numpy.dot(xloc.T, right).T
        else:
            xloc *= right
        assert uloc.shape == xloc.shape
        return xloc

    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Uniform().pdf([-0.5, 0.5, 1.5, 2.5]))
            [0. 1. 0. 0.]
            >>> print(Mul(chaospy.Uniform(), 2).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 0.5 0. ]
            >>> print(Mul(2, chaospy.Uniform()).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 0.5 0. ]
            >>> print(Mul(1, 1.5).pdf([-0.5, 0.5, 1.5, 2.5])) # Dirac logic
            [ 0.  0. inf  0.]
            >>> dist = chaospy.Mul([2, 1], chaospy.Iid(chaospy.Uniform(), 2))
            >>> print(dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [0.5 0.5 0. ]
            >>> dist = chaospy.Mul(chaospy.Iid(chaospy.Uniform(), 2), [1, 2])
            >>> print(dist.pdf([[0.5, 0.6, 1.5], [0.5, 0.6, 1.5]]))
            [0.5 0.5 0. ]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return numpy.inf

        else:
            if self.matrix:
                Ci = numpy.linalg.inv(left)
                xloc = numpy.dot(Ci, xloc)

            else:
                left = (numpy.asfarray(left).T+numpy.zeros(xloc.shape).T).T
                valids = left != 0
                xloc.T[valids.T] = xloc.T[valids.T]/left.T[valids.T]

            pdf = evaluation.evaluate_density(right, xloc, cache=cache)
            if self.matrix:
                pdf = numpy.dot(Ci, pdf)
            else:
                pdf.T[valids.T] /= left.T[valids.T]
            return pdf

        if self.matrix:
            Ci = numpy.linalg.inv(right)
            xloc = numpy.dot(xloc.T, Ci).T
        else:
            right = (numpy.asfarray(right).T+numpy.zeros(xloc.shape).T).T
            valids = right != 0
            xloc.T[valids.T] = xloc.T[valids.T]/right.T[valids.T]
            xloc.T[~valids.T] = numpy.inf

        pdf = evaluation.evaluate_density(left, xloc, cache=cache)
        if self.matrix:
            pdf = numpy.dot(pdf.T, Ci).T
        else:
            pdf.T[valids.T] /= right.T[valids.T]
        assert pdf.shape == xloc.shape
        return pdf

    def _mom(self, key, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> print(numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4))
            [1.     0.5    0.3333 0.25  ]
            >>> print(numpy.around(Mul(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4))
            [1.     1.     1.3333 2.    ]
            >>> print(numpy.around(Mul(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [1.     1.     1.3333 2.    ]
            >>> print(numpy.around(Mul(chaospy.Uniform(), chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [1.     0.25   0.1111 0.0625]
            >>> print(numpy.around(Mul(2, 2).mom([0, 1, 2, 3]), 4))
            [ 1.  4. 16. 64.]
        """
        if evaluation.get_dependencies(left, right):
            raise evaluation.DependencyError(
                "sum of dependent distributions not feasible: "
                "{} and {}".format(left, right)
            )

        if isinstance(left, Dist):
            left = evaluation.evaluate_moment(left, key, cache=cache)
        else:
            left = (numpy.array(left).T**key).T
        if isinstance(right, Dist):
            right = evaluation.evaluate_moment(right, key, cache=cache)
        else:
            right = (numpy.array(right).T**key).T
        return numpy.sum(left*right)

    def _ttr(self, kloc, left, right, cache):
        """Three terms recursion coefficients."""
        if isinstance(left, Dist) and isinstance(right, Dist):
            raise evaluation.DependencyError(
                "product of distributions not feasible: "
                "{} and {}".format(left, right)
            )
        elif isinstance(right, Dist):
            left, right = right, left

        coeff0, coeff1 = evaluation.evaluate_recurrence_coefficients(
            left, kloc, cache=cache)
        right = numpy.asarray(right)
        return coeff0*right, coeff1*right*right

    def __str__(self):
        if self._repr is not None:
            return super(Mul, self).__str__()
        return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                ", " + str(self.prm["right"]) + ")")

    def __len__(self):
        try:
            return len(self.prm["left"])
        except TypeError:
            return 1

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_inverse_cache(self.prm["left"], cache)
        right = evaluation.get_inverse_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left*right
        return self


def mul(left, right):
    """
    Distribution multiplication.

    Args:
        left (Dist, numpy.ndarray) : left hand side.
        right (Dist, numpy.ndarray) : right hand side.
    """
    from .mv_mul import MvMul
    length = max(left, right)
    if length == 1:
        return Mul(left, right)
    return MvMul(left, right)



"""Wald distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class wald(Dist):
    """Wald distribution."""

    def __init__(self, mu):
        Dist.__init__(self, mu=mu)

    def _pdf(self, x, mu):
        out = numpy.zeros(x.shape)
        indices = x > 0
        out[indices] = 1.0/numpy.sqrt(2*numpy.pi*x[indices])
        out[indices] *= numpy.exp(-(1-mu*x[indices])**2.0 / (2*x[indices]*mu**2.0))
        return out

    def _cdf(self, x, mu):
        trm1 = 1./mu - x
        trm2 = 1./mu + x
        isqx = numpy.tile(numpy.inf, x.shape)
        indices = x > 0
        isqx[indices] = 1./numpy.sqrt(x[indices])
        out = 1.-special.ndtr(isqx*trm1)
        out -= numpy.exp(2.0/mu)*special.ndtr(-isqx*trm2)
        return out

    def _bnd(self, x, mu):
        return 0.0, 10**10


class Wald(Add):
    """
    Wald distribution.

    Reciprocal inverse Gaussian distribution.

    Args:
        mu (float, Dist) : Mean of the normal distribution
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Wald(2, 2, 2)
        >>> print(distribution)
        Wald(mu=2, scale=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.7154 3.45   4.5777 6.6903]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.3242 0.2262 0.138  0.063 ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 4.9997  2.4662 11.3302  3.848 ]
    """

    def __init__(self, mu=1, scale=1, shift=0):
        self._repr = {"mu": mu, "scale": scale, "shift": shift}
        Add.__init__(self, left=wald(mu)*scale, right=shift)



"""
Uncertainty Quantification Toolbox
==================================

This module contains tools for performing uncertainty quantification of models.
"""
import logging
import os

import chaospy.bertran
import chaospy.chol
import chaospy.descriptives
import chaospy.distributions
import chaospy.orthogonal
import chaospy.poly
import chaospy.spectral
import chaospy.quadrature
import chaospy.saltelli
import chaospy.regression

from chaospy.distributions import *
from chaospy.orthogonal import *
from chaospy.poly import *
from chaospy.spectral import *
from chaospy.quadrature import *
from chaospy.saltelli import *
from chaospy.descriptives import *
from chaospy.regression import *
from chaospy.external import *

LOGPATH = os.environ.get("CHAOSPY_LOGPATH", os.devnull)
logging.basicConfig(level=logging.DEBUG, filename=LOGPATH, filemode="w")
streamer = logging.StreamHandler()
streamer.setLevel(logging.WARNING)
logger = logging.getLogger(__name__)
logger.addHandler(streamer)



"""Mielke's beta-kappa distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class mielke(Dist):
    """Mielke's beta-kappa distribution."""

    def __init__(self, k, s):
        Dist.__init__(self, k=k, s=s)

    def _pdf(self, x, k, s):
        return k*x**(k-1.0) / (1.0+x**s)**(1.0+k*1.0/s)

    def _cdf(self, x, k, s):
        return x**k / (1.0+x**s)**(k*1.0/s)

    def _ppf(self, q, k, s):
        qsk = pow(q,s*1.0/k)
        return pow(qsk/(1.0-qsk),1.0/s)

    def _bnd(self, x, k, s):
        return 0.0, self._ppf(1-1e-10, k, s)


class Mielke(Add):
    """
    Mielke's beta-kappa distribution

    Args:
        kappa (float, Dist) : First shape parameter
        expo (float, Dist) : Second shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Mielke(2, 0.5, 2)
        >>> print(distribution)
        Mielke(expo=0.5, kappa=2, scale=2, shift=0)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [  6.2633  20.0195  55.867  175.731  919.6095]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0192 0.008  0.0028 0.0007 0.0001]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.58937400e+02 3.88830000e+00 1.21490448e+04 4.99808000e+01]
    """

    def __init__(self, kappa=1, expo=1, scale=1, shift=0):
        self._repr = {
            "kappa": kappa, "expo": expo, "scale": scale, "shift": shift}
        Add.__init__(self, left=mielke(kappa, expo)*scale, right=shift)



"""
Ensure arguments in costructor is correct.
"""

import numpy as np

import chaospy.poly


def ensure_shape(core, shape, shape_):
    """Ensure shape is correct."""
    core = core.copy()
    if shape is None:
        shape = shape_
    elif isinstance(shape, int):
        shape = (shape,)

    if tuple(shape) == tuple(shape_):
        return core, shape

    ones = np.ones(shape, dtype=int)
    for key, val in core.items():
        core[key] = val*ones

    return core, shape


def ensure_dtype(core, dtype, dtype_):
    """Ensure dtype is correct."""
    core = core.copy()
    if dtype is None:
        dtype = dtype_

    if dtype_ == dtype:
        return core, dtype

    for key, val in {
            int: chaospy.poly.typing.asint,
            float: chaospy.poly.typing.asfloat,
            np.float32: chaospy.poly.typing.asfloat,
            np.float64: chaospy.poly.typing.asfloat,
    }.items():

        if dtype == key:
            converter = val
            break
    else:
        raise ValueError("dtype not recognised (%s)" % str(dtype))

    for key, val in core.items():
        core[key] = converter(val)
    return core, dtype


def ensure_dim(core, dim, dim_):
    """Ensure that dim is correct."""
    if dim is None:
        dim = dim_
    if not dim:
        return core, 1
    if dim_ == dim:
        return core, int(dim)

    if dim > dim_:
        key_convert = lambda vari: vari[:dim_]
    else:
        key_convert = lambda vari: vari + (0,)*(dim-dim_)

    new_core = {}
    for key, val in core.items():
        key_ = key_convert(key)
        if key_ in new_core:
            new_core[key_] += val
        else:
            new_core[key_] = val

    return new_core, int(dim)



"""Wrapped Cauchy distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class wrapped_cauchy(Dist):
    """Wrapped Cauchy distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return (1.0-c*c)/(2*numpy.pi*(1+c*c-2*c*numpy.cos(x)))

    def _cdf(self, x, c):
        output = 0.0*x
        val = (1.0+c)/(1.0-c)
        c1 = x<numpy.pi
        c2 = 1-c1

        xn = numpy.extract(c2,x)
        if (any(xn)):
            valn = numpy.extract(c2, numpy.ones_like(x)*val)
            xn = 2*numpy.pi - xn
            yn = numpy.tan(xn/2.0)
            on = 1.0-1.0/numpy.pi*numpy.arctan(valn*yn)
            numpy.place(output, c2, on)

        xp = numpy.extract(c1,x)
        if (any(xp)):
            valp = numpy.extract(c1, numpy.ones_like(x)*val)
            yp = numpy.tan(xp/2.0)
            op = 1.0/numpy.pi*numpy.arctan(valp*yp)
            numpy.place(output, c1, op)

        return output

    def _ppf(self, q, c):
        val = (1.0-c)/(1.0+c)
        rcq = 2*numpy.arctan(val*numpy.tan(numpy.pi*q))
        rcmq = 2*numpy.pi-2*numpy.arctan(val*numpy.tan(numpy.pi*(1-q)))
        return numpy.where(q < 1.0/2, rcq, rcmq)

    def _bnd(self, x, c):
        return 0.0, 2*numpy.pi


class WrappedCauchy(Add):
    """
    Wrapped Cauchy distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.WrappedCauchy(0.8, 4, 6)
        >>> print(distribution)
        WrappedCauchy(scale=4, shape=0.8, shift=6)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 6.5125  7.521  18.5664 29.6117 30.6202]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.2697 0.0928 0.0044 0.0928 0.2697]
        >>> print(numpy.around(distribution.sample(4), 4))
        [29.4606  6.3357 30.9928 14.8313]
        >>> print(numpy.around(distribution.mom(1), 4))
        18.5664
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 18.5664  18.5664  18.5664]
         [121.4247  12.6649  47.0276]]
    """

    def __init__(self, shape=0.5, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=wrapped_cauchy(shape)*scale, right=shift)



"""Kumaraswswamy's double bounded distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class kumaraswamy(Dist):
    """Kumaraswswamy's double bounded distribution."""

    def __init__(self, a=1, b=1):
        assert numpy.all(a > 0) and numpy.all(b > 0)
        Dist.__init__(self, a=a, b=b)

    def _pdf(self, x, a, b):
        return a*b*x**(a-1)*(1-x**a)**(b-1)

    def _cdf(self, x, a, b):
        return 1-(1-x**a)**b

    def _ppf(self, q, a, b):
        return (1-(1-q)**(1./b))**(1./a)

    def _mom(self, k, a, b):
        return b*special.gamma(1+k*1./a)*special.gamma(b)/\
                special.gamma(1+b+k*1./a)

    def _bnd(self, x, a, b):
        return 0,1


class Kumaraswamy(Add):
    """
    Kumaraswamy's double bounded distribution

    Args:
        alpha (float, Dist): First shape parameter, alpha > 0
        beta (float, Dist): Second shape parameter, b > 0
        lower (float, Dist): Lower threshold
        upper (float, Dist): Upper threshold

    Examples:
        >>> distribution = chaospy.Kumaraswamy(2, 2, 2, 3)
        >>> print(distribution)
        Kumaraswamy(alpha=2, beta=2, lower=2, upper=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.3249 2.4748 2.6063 2.7435]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.1625 1.471  1.5337 1.33  ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.6414 2.2434 2.8815 2.5295]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.5333
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.5056 2.5018 2.5008]
         [0.0489 0.0569 0.0595]]
    """

    def __init__(self, alpha, beta, lower=0, upper=1):
        self._repr = {
            "alpha": alpha, "beta": beta, "lower": lower, "upper": upper}
        Add.__init__(
            self, left=kumaraswamy(alpha, beta)*(upper-lower), right=lower)



"""
Even though orthogonal polynomials created using three terms recursion is
the recommended approach as it is the most numerical stable method, it can not
be used directly on stochastically dependent random variables. On method that
bypasses this problem is Cholesky decomposition method.

Cholesky exploits the fact that except for polynomials that are constant, there
is a equivalence between orthogonality and uncorrelated. Decorrelation can be
achieved in a simple few steps:

1. Start with any linear independent basis of polynomials :math:`P`.
2. Temporarily remove any constants, if there are any.
3. Construct the covariance matrix :math`C=Cov(P)` of all polynomials.
4. Decompose covariance matrix into two parts :math:`C = L^T L` using Cholesky
   decomposition.
5. Multiply the vector of polynomials with the inverse decomposition:
   :math:`Q = P L^{-1}`
6. If a constant was removed, subtract the mean from the vector :math`Q=Q-E[Q]`
   before adding the constant back into the expansion.

This should work in theory, but in practice step 4 is known to be numerically
unstable. To this end, this step can to a certain degree be regularized using
various methods. To this end a few modified Cholesky decompositions are
available in ``chaospy``.
"""
import numpy
import chaospy


def orth_chol(order, dist, normed=True, sort="G", cross_truncation=1., **kws):
    """
    Create orthogonal polynomial expansion from Cholesky decomposition.

    Args:
        order (int):
            Order of polynomial expansion
        dist (Dist):
            Distribution space where polynomials are orthogonal
        normed (bool):
            If True orthonormal polynomials will be used instead of monic.
        sort (str):
            Ordering argument passed to poly.basis.  If custom basis is used,
            argument is ignored.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Examples:
        >>> Z = chaospy.Normal()
        >>> print(chaospy.around(chaospy.orth_chol(3, Z), 4))
        [1.0, q0, 0.7071q0^2-0.7071, 0.4082q0^3-1.2247q0]
    """
    dim = len(dist)
    basis = chaospy.poly.basis(
        start=1, stop=order, dim=dim, sort=sort,
        cross_truncation=cross_truncation,
    )
    length = len(basis)

    cholmat = chaospy.chol.gill_king(chaospy.descriptives.Cov(basis, dist))
    cholmat_inv = numpy.linalg.inv(cholmat.T).T
    if not normed:
        diag_mesh = numpy.repeat(numpy.diag(cholmat_inv), len(cholmat_inv))
        cholmat_inv /= diag_mesh.reshape(cholmat_inv.shape)

    coefs = numpy.empty((length+1, length+1))

    coefs[1:, 1:] = cholmat_inv
    coefs[0, 0] = 1
    coefs[0, 1:] = 0

    expected = -numpy.sum(
        cholmat_inv*chaospy.descriptives.E(basis, dist, **kws), -1)
    coefs[1:, 0] = expected

    coefs = coefs.T

    out = {}
    out[(0,)*dim] = coefs[0]
    for idx in range(length):
        index = basis[idx].keys[0]
        out[index] = coefs[idx+1]

    polynomials = chaospy.poly.Poly(out, dim, coefs.shape[1:], float)

    return polynomials


def norm(order, dist, orth=None):

    dim = len(dist)
    try:
        if dim>1:
            norms = numpy.array([norm(order+1, D) for D in dist])
            Is = chaospy.bertran.bindex(order, dim)
            out = numpy.ones(len(Is))

            for i in range(len(Is)):
                index = Is[i]
                for j in range(dim):
                    if index[j]:
                        out[i] *= norms[j, index[j]]
            return out

        K = range(1,order+1)
        ttr = [1.] + [dist.ttr(k)[1] for k in K]
        return numpy.cumprod(ttr)

    except NotImplementedError:

        if orth is None:
            orth = orth_chol(order, dist)
        return chaospy.descriptives.E(orth**2, dist)



# -*- coding: utf-8 -*-
"""
Fejér proposed two quadrature rules very similar to :ref:`clenshaw_curtis`.
The only difference is that the endpoints are set to zero. That is, Fejér only
used the interior extrema of the Chebyshev polynomials, i.e. the true
stationary points. This makes this a better method for performing quadrature on
infinite intervals, as the evaluation does not contain illegal values.

Example usage
-------------

The first few orders with linear growth rule::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3]:
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="fejer")
    ...     print(order, numpy.around(X, 3), numpy.around(W, 3))
    0 [[0.5]] [1.]
    1 [[0.25 0.75]] [0.5 0.5]
    2 [[0.146 0.5   0.854]] [0.286 0.429 0.286]
    3 [[0.095 0.345 0.655 0.905]] [0.188 0.312 0.312 0.188]

The first few orders with exponential growth rule::

    >>> for order in [0, 1, 2]:  # doctest: +NORMALIZE_WHITESPACE
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="fejer", growth=True)
    ...     print(order, numpy.around(X, 2), numpy.around(W, 2))
    0 [[0.5]] [1.]
    1 [[0.15 0.5  0.85]] [0.29 0.43 0.29]
    2 [[0.04 0.15 0.31 0.5  0.69 0.85 0.96]]
        [0.07 0.14 0.18 0.2  0.18 0.14 0.07]

Applying the rule using Smolyak sparse grid::

    >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
    >>> X, W = chaospy.generate_quadrature(
    ...     2, distribution, rule="fejer", growth=True, sparse=True)
    >>> print(numpy.around(X, 3))  # doctest: +NORMALIZE_WHITESPACE
    [[0.038 0.146 0.146 0.146 0.309 0.5   0.5   0.5   0.5
      0.5   0.5   0.5   0.691 0.854 0.854 0.854 0.962]
     [0.5   0.146 0.5   0.854 0.5   0.038 0.146 0.309 0.5
      0.691 0.854 0.962 0.5   0.146 0.5   0.854 0.5  ]]
    >>> print(numpy.around(W, 3))  # doctest: +NORMALIZE_WHITESPACE
    [ 0.074  0.082 -0.021  0.082  0.184  0.074 -0.021  0.184 -0.273
      0.184 -0.021  0.074  0.184  0.082 -0.021  0.082  0.074]
"""
from __future__ import division, print_function

import numpy

from .combine import combine_quadrature


def quad_fejer(order, domain=(0, 1), growth=False):
    """
    Generate the quadrature abscissas and weights in Fejer quadrature.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Dist, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = quad_fejer(3, (0, 1))
        >>> print(numpy.around(abscissas, 4))
        [[0.0955 0.3455 0.6545 0.9045]]
        >>> print(numpy.around(weights, 4))
        [0.1804 0.2996 0.2996 0.1804]
    """
    from ..distributions.baseclass import Dist
    if isinstance(domain, Dist):
        abscissas, weights = quad_fejer(
            order, domain.range(), growth)
        weights *= domain.pdf(abscissas).flatten()
        weights /= numpy.sum(weights)
        return abscissas, weights

    order = numpy.asarray(order, dtype=int).flatten()
    lower, upper = numpy.array(domain)
    lower = numpy.asarray(lower).flatten()
    upper = numpy.asarray(upper).flatten()

    dim = max(lower.size, upper.size, order.size)

    order = numpy.ones(dim, dtype=int)*order
    lower = numpy.ones(dim)*lower
    upper = numpy.ones(dim)*upper

    if growth:
        order = numpy.where(order > 0, 2**(order+1)-2, 0)

    abscissas, weights = zip(*[_fejer(order_) for order_ in order])

    return combine_quadrature(abscissas, weights, (lower, upper))


def _fejer(order):
    r"""
    Backend method.

    Examples:
        >>> abscissas, weights = _fejer(0)
        >>> print(abscissas)
        [0.5]
        >>> print(weights)
        [1.]
        >>> abscissas, weights = _fejer(1)
        >>> print(abscissas)
        [0.25 0.75]
        >>> print(weights)
        [0.44444444 0.44444444]
        >>> abscissas, weights = _fejer(2)
        >>> print(abscissas)
        [0.14644661 0.5        0.85355339]
        >>> print(weights)
        [0.26666667 0.4        0.26666667]
        >>> abscissas, weights = _fejer(3)
        >>> print(abscissas)
        [0.0954915 0.3454915 0.6545085 0.9045085]
        >>> print(weights)
        [0.18037152 0.29962848 0.29962848 0.18037152]
        >>> abscissas, weights = _fejer(4)
        >>> print(abscissas)
        [0.0669873 0.25      0.5       0.75      0.9330127]
        >>> print(weights)
        [0.12698413 0.22857143 0.26031746 0.22857143 0.12698413]
        >>> abscissas, weights = _fejer(5)
        >>> print(abscissas)
        [0.04951557 0.1882551  0.38873953 0.61126047 0.8117449  0.95048443]
        >>> print(weights)
        [0.0950705  0.17612121 0.2186042  0.2186042  0.17612121 0.0950705 ]
    """
    order = int(order)
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])

    order += 2

    theta = (order-numpy.arange(order+1))*numpy.pi/order
    abscisas = 0.5*numpy.cos(theta) + 0.5

    idx, idy = numpy.mgrid[:order+1, :order//2]
    weights = 2*numpy.cos(2*(idy+1)*theta[idx])/(4*idy*(idy+2)+3)
    if order % 2 == 0:
        weights[:, -1] *= 0.5
    weights = (1-numpy.sum(weights, -1)) / order

    return abscisas[1:-1], weights[1:-1]



r"""
Most integration problems when dealing with polynomial chaos expansion comes
with a weight function :math:`p(x)` which happens to be the probability density
function. Gaussian quadrature creates weights and abscissas that are tailored
to be optimal with the inclusion of a weight function. It is therefore not one
method, but a collection of methods, each tailored to different probability
density functions.

In ``chaospy`` Gaussian quadrature is a functionality attached to each
probability distribution. This means that instead of explicitly supporting
a list of quadrature rules, all rules are supported through the capability of
the distribution implementation. For common distribution, this means that the
quadrature rules are calculated analytically using Stieltjes method on known
three terms recursion coefficients, and using those to create quadrature node
using the e.g. discretized Stieltjes algorithm.

For example for the tailored quadrature rules defined above:

* Gauss-Hermit quadrature is tailored to the normal (Gaussian) distribution::

    >>> distribution = chaospy.Normal(0, 1)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[-3.3243 -1.8892 -0.6167  0.6167  1.8892  3.3243]]
    >>> print(numpy.around(W, 4))
    [0.0026 0.0886 0.4088 0.4088 0.0886 0.0026]

* Gauss-Legendre quadrature is tailored to the Uniform distributions::

    >>> distribution = chaospy.Uniform(-1, 1)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[-0.9325 -0.6612 -0.2386  0.2386  0.6612  0.9325]]
    >>> print(numpy.around(W, 4))
    [0.0857 0.1804 0.234  0.234  0.1804 0.0857]

* Gauss-Jacobi quadrature is tailored to the Beta distribution::

    >>> distribution = chaospy.Beta(2, 4, lower=-1, upper=1)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[-0.8969 -0.6679 -0.3448  0.0289  0.4029  0.7279]]
    >>> print(numpy.around(W, 4))
    [0.0749 0.272  0.355  0.2253 0.0667 0.0062]

* Gauss-Laguerre quadrature is tailored to the Exponential distribution::

    >>> distribution = chaospy.Exponential()
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[ 0.2228  1.1889  2.9927  5.7751  9.8375 15.9829]]
    >>> print(numpy.around(W, 4))
    [4.590e-01 4.170e-01 1.134e-01 1.040e-02 3.000e-04 0.000e+00]

* Generalized Gauss-Laguerre quadrature is tailored to the Gamma distribution::

    >>> distribution = chaospy.Gamma(2, 4)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[ 2.1107  7.1852 15.5066 27.6753 44.9384 70.5839]]
    >>> print(numpy.around(W, 4))
    [0.2777 0.4939 0.203  0.0247 0.0008 0.    ]

For uncommon distributions an analytical Stieltjes method can not be performed
as the distribution does not provide three terms recursion coefficients. In
this scenario, the discretized counterpart is used instead as an approximation.
For example, to mention a few:

* The Triangle distribution::

    >>> distribution = chaospy.Triangle(-1, 0, 1)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[-0.8657 -0.5766 -0.1943  0.1943  0.5766  0.8657]]
    >>> print(numpy.around(W, 4))
    [0.0295 0.1475 0.323  0.323  0.1475 0.0295]

* The Laplace distribution::

    >>> distribution = chaospy.Laplace(0, 1)
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[-10.5577  -4.6629  -1.0424   1.0424   4.6629  10.5577]]
    >>> print(numpy.around(W, 4))
    [1.000e-04 2.160e-02 4.783e-01 4.783e-01 2.160e-02 1.000e-04]

* The Weibull distribution::

    >>> distribution = chaospy.Weibull()
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[ 0.2228  1.1886  2.9918  5.7731  9.8334 15.9737]]
    >>> print(numpy.around(W, 4))
    [4.589e-01 4.170e-01 1.134e-01 1.040e-02 3.000e-04 0.000e+00]

* The Rayleigh distribution::

    >>> distribution = chaospy.Rayleigh()
    >>> X, W = chaospy.generate_quadrature(
    ...     5, distribution, rule="gaussian")
    >>> print(numpy.around(X, 4))
    [[0.2473 0.7687 1.4795 2.3314 3.3228 4.5295]]
    >>> print(numpy.around(W, 4))
    [9.600e-02 3.591e-01 3.891e-01 1.412e-01 1.430e-02 2.000e-04]
"""
from .recurrence import (
    construct_recurrence_coefficients, coefficients_to_quadrature)
from .combine import combine_quadrature


def quad_gaussian(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generating Gaussian quadrature by first generating so called *three terms
    recurrence* coefficients using one various different algorithms. The
    coefficients are them converted to abscissas and weights by constructing
    lower banded Jakobi matrix and calculating eigenvalues and eigenvectors,
    which can be directly translated to abscissas and weights. Construction of
    the coefficients is potentially numerically unstable, there for multiple
    algorithms exists.

    Args:
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
        order (int):
            The order of the quadrature.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Gaussian quadrature abscissas and weights with shapes
            ``(len(dist), order+1)`` and ``(order+1,)`` respectively.

    Raises:
        NotImplementedError:
            In the case of recurrence algorithm ``analytical``, error is raised
            if the distribution does not implement the three terms recurrence
            algorithm analytically.
        numpy.linalg.LinAlgError:
            For non-canonical random variables, the construction might fail
            because of illegal numerical operations.

    Examples:
        >>> distribution = chaospy.Normal(0, 1)
        >>> abscissas, weights = chaospy.quad_gaussian(
        ...     5, distribution, recurrence_algorithm="stieltjes")
        >>> print(numpy.around(abscissas, 4))
        [[-3.3243 -1.8892 -0.6167  0.6167  1.8892  3.3243]]
        >>> print(numpy.around(weights, 4))
        [0.0026 0.0886 0.4088 0.4088 0.0886 0.0026]
        >>> distribution = chaospy.J(chaospy.Uniform(), chaospy.Normal())
        >>> abscissas, weights = chaospy.quad_gaussian(
        ...     2, distribution, recurrence_algorithm="chebyshev")
        >>> print(numpy.around(abscissas, 3))
        [[ 0.113  0.113  0.113  0.5    0.5    0.5    0.887  0.887  0.887]
         [-1.732  0.     1.732 -1.732  0.     1.732 -1.732  0.     1.732]]
        >>> print(numpy.around(weights, 4))
        [0.0463 0.1852 0.0463 0.0741 0.2963 0.0741 0.0463 0.1852 0.0463]
    """
    coefficients = construct_recurrence_coefficients(
        order, dist, rule, accuracy, recurrence_algorithm)
    abscissas, weights = coefficients_to_quadrature(coefficients)
    return combine_quadrature(abscissas, weights)



"""
Laja quadrature is a newer method for performing quadrature in stochastical
problems. The method is described in a `journal paper`_ by Narayan and Jakeman.

.. _journal paper: https://arxiv.org/pdf/1404.5663.pdf

Example usage
-------------

The first few orders::

    >>> distribution = chaospy.Uniform(0, 1)
    >>> for order in [0, 1, 2, 3, 4]:
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="leja")
    ...     print(order, numpy.around(abscissas, 3),
    ...           numpy.around(weights, 3))
    0 [[0.5]] [1.]
    1 [[0.5 1. ]] [1. 0.]
    2 [[0.  0.5 1. ]] [0.167 0.667 0.167]
    3 [[0.    0.5   0.789 1.   ]] [0.167 0.667 0.    0.167]
    4 [[0.    0.171 0.5   0.789 1.   ]] [0.043 0.289 0.316 0.28  0.072]
"""
from __future__ import print_function

import numpy
from scipy.optimize import fminbound

from .combine import combine
from .recurrence import analytical_stieljes, discretized_stieltjes


def quad_leja(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generate Leja quadrature node.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscisas, weights = quad_leja(3, chaospy.Normal(0, 1))
        >>> print(numpy.around(abscisas, 4))
        [[-2.7173 -1.4142  0.      1.7635]]
        >>> print(numpy.around(weights, 4))
        [0.022  0.1629 0.6506 0.1645]
    """
    from chaospy.distributions import evaluation

    if len(dist) > 1:
        if evaluation.get_dependencies(*list(dist)):
            raise evaluation.DependencyError(
                "Leja quadrature do not supper distribution with dependencies.")
        if isinstance(order, int):
            out = [quad_leja(order, _) for _ in dist]
        else:
            out = [quad_leja(order[_], dist[_]) for _ in range(len(dist))]

        abscissas = [_[0][0] for _ in out]
        weights = [_[1] for _ in out]
        abscissas = combine(abscissas).T
        weights = combine(weights)
        weights = numpy.prod(weights, -1)

        return abscissas, weights

    lower, upper = dist.range()
    abscissas = [lower, dist.mom(1), upper]
    for _ in range(int(order)):

        def objective(abscissas_):
            """Local objective function."""
            return -numpy.sqrt(dist.pdf(abscissas_))*numpy.prod(
                numpy.abs(abscissas[1:-1]-abscissas_))

        opts, vals = zip(
            *[fminbound(
                objective, abscissas[idx], abscissas[idx+1], full_output=1)[:2]
              for idx in range(len(abscissas)-1)]
        )
        index = numpy.argmin(vals)
        abscissas.insert(index+1, opts[index])

    abscissas = numpy.asfarray(abscissas).flatten()[1:-1]
    weights = create_weights(
        abscissas, dist, rule, accuracy, recurrence_algorithm)
    abscissas = abscissas.reshape(1, abscissas.size)

    return numpy.asfarray(abscissas), numpy.asfarray(weights)


def create_weights(
        nodes,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """Create weights for the Laja method."""
    try:
        _, poly, _ = analytical_stieljes(len(nodes)-1, dist)
    except NotImplementedError:
        from .frontend import generate_quadrature
        abscissas, weights = generate_quadrature(
            order=accuracy, dist=dist, rule=rule,
            recurrence_algorithm=recurrence_algorithm,
        )
        _, poly, _ = discretized_stieltjes(
            len(nodes)-1, abscissas, weights)
    weights = numpy.linalg.inv(poly(nodes))
    return weights[:, 0]



"""Normal (Gaussian) probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class normal(Dist):
    """Standard normal distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)

    def _cdf(self, x):
        return special.ndtr(x)

    def _ppf(self, x):
        return special.ndtri(x)

    def _mom(self, k):
        return .5*special.factorial2(k-1)*(1+(-1)**k)

    def _ttr(self, n):
        return 0., 1.*n

    def _bnd(self, x):
        return -7.5, 7.5


class Normal(Add):
    R"""
    Normal (Gaussian) distribution

    Args:
        mu (float, Dist) : Mean of the distribution.
        sigma (float, Dist) : Standard deviation.  sigma > 0

    Examples:
        >>> distribution = chaospy.Normal(2, 2)
        >>> print(distribution)
        Normal(mu=2, sigma=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.3168 1.4933 2.5067 3.6832]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.14   0.1932 0.1932 0.14  ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.7901 -0.4006  5.2952  1.9107]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 2.  2.  2.]
         [ 4.  8. 12.]]
    """

    def __init__(self, mu=0, sigma=1):
        self._repr = {"mu": mu, "sigma": sigma}
        Add.__init__(self, left=normal()*sigma, right=mu)



"""Chi distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class chi(Dist):
    """Chi distribution."""

    def __init__(self, df=1):
        Dist.__init__(self, df=df)

    def _pdf(self, x, df):
        return x**(df-1.)*numpy.exp(-x*x*0.5)/(2.0)**(df*0.5-1)\
                /special.gamma(df*0.5)

    def _cdf(self, x, df):
        return special.gammainc(df*0.5,0.5*x*x)

    def _ppf(self, q, df):
        return numpy.sqrt(2*special.gammaincinv(df*0.5,q))

    def _bnd(self, x, df):
        return 0, self._ppf(1-1e-10, df)

    def _mom(self, k, df):
        return 2**(.5*k)*special.gamma(.5*(df+k))\
                /special.gamma(.5*df)


class Chi(Add):
    """
    Chi distribution.

    Args:
        df (float, Dist) : Degrees of freedom
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Chi(2, 4, 1)
        >>> print(distribution)
        Chi(df=2, scale=4, shift=1)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 1.      4.0341  5.7096  7.6604 28.1446]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.1422 0.1472 0.1041 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 6.8244  2.9773 10.8003  5.5892]
        >>> print(numpy.around(distribution.mom(1), 4))
        6.0133
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 7.6671  9.0688 10.2809]
         [ 6.8673 12.6824 18.2126]]
    """

    def __init__(self, df=1, scale=1, shift=0):
        self._repr = {"df": df, "scale": scale, "shift": shift}
        Add.__init__(self, left=chi(df)*scale, right=shift)


class Maxwell(Add):
    """
    Maxwell-Boltzmann distribution
    Chi distribution with 3 degrees of freedom

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Maxwell(2, 3)
        >>> print(distribution)
        Maxwell(scale=2, shift=3)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 3.      5.2023  6.0763  7.0538 17.0772]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.2638 0.2892 0.2101 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [6.6381 4.6119 8.5955 6.015 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        6.1915
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[6.8457 7.4421 7.9834]
         [1.8141 3.3964 4.8716]]
    """

    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=chi(3)*scale, right=shift)


class Rayleigh(Add):
    """
    Rayleigh distribution

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Rayleigh(2, 3)
        >>> print(distribution)
        Rayleigh(scale=2, shift=3)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 3.      4.5171  5.3548  6.3302 16.5723]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.2844 0.2944 0.2081 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.9122 3.9886 7.9001 5.2946]
        >>> print(numpy.around(distribution.mom(1), 4))
        5.5066
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[6.3336 7.0344 7.6405]
         [1.7168 3.1706 4.5532]]
    """
    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=chi(2)*scale, right=shift)



"""Log-gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class log_gamma(Dist):
    """Log-gamma distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.exp(c*x-numpy.exp(x)-special.gammaln(c))

    def _cdf(self, x, c):
        return special.gammainc(c, numpy.exp(x))

    def _ppf(self, q, c):
        return numpy.log(special.gammaincinv(c,q))

    def _bnd(self, x, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class LogGamma(Add):
    """
    Log-gamma distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.LogGamma(2, 2, 1)
        >>> print(distribution)
        LogGamma(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.6138 1.639  2.4085 3.1934]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.149  0.2392 0.2706 0.2245]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.6074 -0.0932  4.1166  1.9675]
        >>> print(numpy.around(distribution.mom(1), 4))
        1.8456
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 0.5924 -0.8925 -2.5379]
         [ 2.5797  6.6525 12.5947]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=log_gamma(shape)*scale, right=shift)


Loggamma = deprecation_warning(LogGamma, "Loggamma")



"""Variance based decomposition on the Nataf Copula."""
import numpy

import chaospy

from ..conditional import E_cond
from ..variance import Var


def Sens_m_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition through the Nataf distribution.

    Generates first order sensitivity indices

    Args:
        order (int):
            Polynomial order used ``orth_ttr``.
        dist (Copula):
            Assumed to be Nataf with independent components
        samples (numpy.ndarray):
            Samples used for evaluation (typically generated from ``dist``.)
        vals (numpy.ndarray): Evaluations of the model for given samples.

    Returns:
        (numpy.ndarray):
            Sensitivity indices with shape ``(len(dist),) + vals.shape[1:]``.
    """
    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = numpy.array(vals)

    cov = trans.prm["C"]
    cov = numpy.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index = [1] + [0]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = numpy.zeros((dim,) + poly.shape)
    out[0] = Var(E_cond(poly, index, marginal, **kws),
                 marginal, **kws)/(V+(V == 0))*(V != 0)


    for i in range(1, dim):

        r = r[1:] + r[:1]
        index = index[-1:] + index[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[i] = Var(E_cond(poly, index, marginal, **kws),
                     marginal, **kws)/(V+(V == 0))*(V != 0)

    return out


def Sens_t_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition thorugh the Nataf distribution.

    Total order sensitivity indices

    Args:
        order (int):
            Polynomial order used ``orth_ttr``.
        dist (Copula):
            Assumed to be Nataf with independent components
        samples (numpy.ndarray):
            Samples used for evaluation (typically generated from ``dist``.)
        vals (numpy.ndarray):
            Evaluations of the model for given samples.

    Returns:
        (numpy.ndarray) :
            Sensitivity indices with shape ``(len(dist),)+vals.shape[1:]``.
    """
    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = nump.array(vals)

    cov = trans.prm["C"]
    cov = nump.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index = [0] + [1]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = nump.zeros((dim,) + poly.shape)
    out[0] = (V-Var(E_cond(poly, index, marginal, **kws),
                    marginal, **kws))/(V+(V == 0))**(V != 0)

    for i in range(1, dim):

        r = r[1:] + r[:1]
        index = index[-1:] + index[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[i] = (V-Var(E_cond(poly, index, marginal, **kws),
                        marginal, **kws))/(V+(V == 0))*(V != 0)

    return out


def Sens_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition through the Nataf distribution.

    Main and total order sensitivity indices

    Args:
        order (int):
            polynomial order used ``orth_ttr``.
        dist (Copula):
            Assumed to be Nataf with independent components
        samples (numpy.ndarray):
            Samples used for evaluation (typically generated from ``dist``.)
        vals (numpy.ndarray):
            Evaluations of the model for given samples.

    Returns:
        (numpy.ndarray):
            Sensitivity indices with shape
            ``kjl``(2, len(dist),)+vals.shape[1:]``. First component is main
            and second is total.
    """
    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = nump.array(vals)

    cov = trans.prm["C"]
    cov = nump.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index0 = [0] + [1]*(dim-1)
    index1 = [1] + [0]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = nump.zeros((2, dim,) + poly.shape)
    out[0, 0] = (V - Var(E_cond(poly, index0, marginal, **kws),
                        marginal, **kws))/(V+(V == 0))**(V != 0)
    out[1, 0] = Var(E_cond(poly, index1, marginal, **kws),
                   marginal, **kws)/(V+(V == 0))*(V != 0)

    for i in range(1, dim):

        r = r[1:] + r[:1]
        index0 = index0[-1:] + index0[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[0, i] = (V-Var(E_cond(poly, index0, marginal, **kws),
                          marginal, **kws))/(V+(V == 0))*(V != 0)
        out[1, i] = Var(E_cond(poly, index1, marginal, **kws),
                       marginal, **kws)/(V+(V == 0))*(V != 0)

    return out[::-1]



"""
Evaluate raw statistical moments.

Example usage
-------------

Define a simple distribution and data::

    >>> class Exponential(chaospy.Dist):
    ...     def _mom(self, k_data, alpha): return alpha**-k_data
    >>> dist = Exponential(alpha=2.)
    >>> k_data = numpy.array([1])

Normal usage::

    >>> print(evaluate_moment(dist, k_data))
    0.5
    >>> print(evaluate_moment(dist, k_data, parameters={"alpha": 1.}))
    1.0

0-th order is always precomputed::

    >>> print(evaluate_moment(dist, numpy.array([0])))
    1.0

The use of cache::

    >>> print(evaluate_moment(dist, k_data, cache={((1,), dist): 5.}))
    5.0

Approximate with the use of density function if moment is missing::

    >>> class Exponential(chaospy.Dist):
    ...     def _cdf(self, x_data, alpha): return 1-numpy.e**(-alpha*x_data)
    ...     def _bnd(self, x_data, alpha): return 0, 50
    ...     def _pdf(self, x_data, alpha): return alpha*numpy.e**(-alpha*x_data)
    >>> dist = Exponential(alpha=2)
    >>> print(numpy.around(evaluate_moment(dist, k_data), 4))
    0.5
"""
import logging

import numpy
from .parameters import load_parameters


def evaluate_moment(
        distribution,
        k_data,
        parameters=None,
        cache=None,
):
    """
    Evaluate raw statistical moments.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        x_data (numpy.ndarray):
            Locations for where evaluate moment of.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The raw statistical moment of ``distribution`` at location ``x_data``
        using parameters ``parameters``.
    """
    logger = logging.getLogger(__name__)
    assert len(k_data) == len(distribution), (
        "distribution %s is not of length %d" % (distribution, len(k_data)))
    assert len(k_data.shape) == 1

    if numpy.all(k_data == 0):
        return 1.

    def cache_key(distribution):
        return (tuple(k_data), distribution)

    if cache is None:
        cache = {}
    else:
        if cache_key(distribution) in cache:
            return cache[cache_key(distribution)]

    from .. import baseclass
    try:
        parameters = load_parameters(
            distribution, "_mom", parameters, cache, cache_key)
        out = distribution._mom(k_data, **parameters)

    except baseclass.StochasticallyDependentError:

        logger.info(
            "Distribution %s has stochastic dependencies; "
            "Approximating moments with quadrature.", distribution)
        from .. import approximation
        out = approximation.approximate_moment(distribution, k_data)

    if isinstance(out, numpy.ndarray):
        out = out.item()

    cache[cache_key(distribution)] = out

    return out



"""Folded Cauchy distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class folded_cauchy(Dist):
    """Folded Cauchy distribution."""

    def __init__(self, c=0):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return 1./(numpy.pi*(1+(x-c)**2)) + 1./(numpy.pi*(1+(x+c)**2))

    def _cdf(self, x, c):
        return (numpy.arctan(x-c) + numpy.arctan(x+c))/numpy.pi

    def _bnd(self, x, c):
        return 0, 10**10


class FoldedCauchy(Add):
    """
    Folded Cauchy distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.FoldedCauchy(3, 2, 1)
        >>> print(distribution)
        FoldedCauchy(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 5.1449  6.708   8.0077 10.6504]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0915 0.1603 0.1306 0.0393]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 8.4584  3.9544 27.8887  7.2135]
    """

    def __init__(self, shape=0, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=folded_cauchy(shape)*scale, right=shift)


Foldcauchy = deprecation_warning(FoldedCauchy, "Foldcauchy")
